/*------------------------------------------------------------------------
 |  FILE            SpaceGameApplet.java
 |  PART OF         SHH SpaceGame - a simple, networked multiplayer game.
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

import java.applet.*;
import java.net.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;

import client.*;
import objects.*;
import gui.*;

/**
 *
 * @author    Sverre H. Huseby
 *            &lt;<a href="mailto:shh@thathost.com">shh@thathost.com</a>&gt;
 * @version   $Id: SpaceGameApplet.java,v 1.5 2003/03/10 21:26:12 sverrehu Exp $
 */
public final class SpaceGameApplet
extends Applet
implements MainFuncProvider, ActionListener {
/*-----------------------------------------------------------------------+
 |  PRIVATE PART                                                         |
 +----------------------------------------------------------------------*/
    private int        boardWidth;
    private int        boardHeight;
    private String     remoteServerName;
    private int        remoteServerPort;
    private String     myName;
    private boolean    gotPlayerName;
    private Board      board;
    private World      world;
    private Client     client;
    private TextField  chatInput;
    private Label      messageLabel;
    private Component  centerComponent;

    private String getParam(String name, String dflt) {
        String value;

        if ((value = getParameter(name)) == null)
            return dflt;
        return value;
    }

    private int getIntParam(String name, int dflt) {
        String value;

        if ((value = getParameter(name)) == null)
            return dflt;
        try {
            return Integer.parseInt(value);
        } catch (NumberFormatException e) {
        }
        return dflt;
    }

    private void showCenter(Component comp) {
        if (centerComponent != null)
            remove(centerComponent);
        if (comp != null)
            add(comp, BorderLayout.CENTER);
        centerComponent = comp;
    }

    private void handleException(Exception e, String msg) {
        msg += ": " + e.getMessage();
        System.err.println(msg);
        messageLabel.setText(msg);
        showCenter(messageLabel);
    }

    private void setupWindowForNameQuery() {
        gotPlayerName = false;

        messageLabel = new Label();
        messageLabel.setAlignment(Label.CENTER);
        messageLabel.setBackground(Color.black);
        messageLabel.setForeground(Color.green);

        setLayout(new BorderLayout());

        chatInput = new TextField();
        chatInput.setBackground(Color.black);
        chatInput.setForeground(Color.green);
        chatInput.addActionListener(this);
        add(chatInput, BorderLayout.SOUTH);

        messageLabel.setText("Please enter your name on the line below");
        showCenter(messageLabel);

        chatInput.requestFocus();
    }

    private void setupWindowForPlaying() {
        world = new World();
        board = new Board(this, world,
                          boardWidth,
                          boardHeight - (chatInput.getSize().height + 1));
        messageLabel.setText("");
        showCenter(board);

        board.requestFocus();
    }

    private void connectToServer(String host, int port) {
        messageLabel.setText("Connecting to " + host + ":" + port + "...");
        try {
            client = new Client(this, host, port, world, board, null);
        } catch (IOException e) {
            handleException(e, "Unable to connect to server");
            return;
        }

        client.start();
        client.startUpdater();
        client.sendLogin(myName);
        client.flush();
    }

    private void disconnectFromServer() {
        if (client != null) {
            client.stopUpdater();
            client.stop();
            client = null;
        }
    }

/*-----------------------------------------------------------------------+
 |  PUBLIC INTERFACE                                                     |
 +----------------------------------------------------------------------*/

    /* Applet ***********************************************************/
    public void init() {
        String host;

        System.out.println("SpaceGame Init");

        /* set up defaults */
        boardWidth = 600;
        boardHeight = 600;
        remoteServerName = "localhost";
        remoteServerPort = 9998;
        host = getCodeBase().getHost();
        if (host == null || host.length() == 0)
            host = remoteServerName;

        boardWidth       = getIntParam("width", boardWidth);
        boardHeight      = getIntParam("height", boardHeight);
        remoteServerName = getParam("host", host);
        remoteServerPort = getIntParam("port", remoteServerPort);
    }

    public void start() {
        System.out.println("SpaceGame Starting");
        setupWindowForNameQuery();
    }

    public void stop() {
        System.out.println("SpaceGame Stopping");
        if (chatInput != null) {
            remove(chatInput);
            chatInput = null;
        }
        showCenter(null);
        disconnectFromServer();
        board = null;
        world = null;
    }

    public void destroy() {
        System.out.println("SpaceGame Unloading");
    }

    /* MainFuncProvider *************************************************/
    /* called by client.Board */
    public void invokeChatLine() {
        if (chatInput != null)
            chatInput.requestFocus();
    }

    /* called by client.Client */
    public void doQuit() {
        stop();
    }

    /* ActionListener ***************************************************/
    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == chatInput) {
            String msg;

            msg = chatInput.getText();
            if (gotPlayerName) {
                /* chat message */
                if (msg.length() > 0 && client != null) {
                    client.sendSay(msg);
                    client.flush();
                }
                chatInput.setText("");
                board.requestFocus();
            } else {
                /* request for player name */
                msg = msg.trim();
                if (msg.length() > 0) {
                    chatInput.setText("");
                    myName = msg;
                    gotPlayerName = true;
                    setupWindowForPlaying();
                    connectToServer(remoteServerName, remoteServerPort);
                }
            }
        }
    }
}
