/*------------------------------------------------------------------------
 |  FILE            Aim.java
 |  WRITTEN BY      Sverre H. Huseby <shh@computas.no>
 +----------------------------------------------------------------------*/

package userclient;

import java.awt.*;        /* Point, etc. */
import java.awt.event.*;  /* keycodes */

import client.*;
import objects.*;

/**
 * A simple example on a "smart" client. This one will help you aim at
 * the closest enemy player. It is not very smart, in that it always
 * aims at the center, even if the enemy is moving.
 *
 * If you turn your ship manually, the aimer will stop aiming, and wait
 * untill a couple of seconds after you stop you manual turning before
 * aiming again. You may disable/enable the aimer by pressing the "Z" key.
 *
 * @author    Sverre H. Huseby
 *            &lt;<A HREF="mailto:shh@computas.no">shh@computas.no</A>&gt;
 * @version   $Id: Aim.java,v 1.4 2001/10/16 18:54:03 sverrehu Exp $
 */
public final class Aim
extends UserDefinedClient {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    /** Number of seconds to delay when temporarily disabled. */
    private static final int delaySeconds = 2;

    private boolean disabled;        /* currently fully disabled? */
    private boolean manualTurnLeft;  /* user currently turning manually? */
    private boolean manualTurnRight;
    private int     delayCount;      /* timer for delays */

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    /* this one is called by the "main system" 15 times a second or so
     * (check getUpdateFrequency() for the accurate number), so this is
     * the method that provides all the logic of the client. */
    protected synchronized final void update() {
	Player me, enemy;
	Point  myLoc, enemyLoc;
	double currentAngle, wantedAngle;
	int    turn;

	/* check if we are temporarily disabled. */
	if (delayCount > 0) {
	    /* we are disabled for a while. count down. */
	    --delayCount;
	    return;
	}
	/* check if we are fully disabled. */
	if (disabled)
	    return;
	if (manualTurnLeft || manualTurnRight) {
	    /* the user is currently turning the ship manually, so we
	     * disable ourselves for a short period of time, to avoid
	     * interferring with the users aims. */
	    delayCount = delaySeconds * getUpdateFrequency();
	    return;
	}

	/* now pick an enemy to aim in at. what a convenience! we have
         * a method that picks it for us! Sverre is a nice guy. */
	if ((enemy = getClosestEnemy()) == null)
	    return;
	me = getMyPlayer();

	/* now we want to aim in on the enemy. the turning interface
	 * allows us to start and stop turning. we need to know which
	 * direction to turn to, and when we are finished
	 * turning. first find the angle we want to end up in. */
	myLoc = me.getLocation();
	enemyLoc = enemy.getLocation();
	wantedAngle = getAngle(myLoc.x, myLoc.y, enemyLoc.x, enemyLoc.y);

	/* so far, so good. now we need to know which way to
         * turn. luckily, we have a method that tells us this based on
         * our current direction, and the direction we want. */
	currentAngle = me.getDirection();
	turn = getBestTurn(currentAngle, wantedAngle);

	/* ready for turning. */
	switch (turn) {
	  case TURN_LEFT:
	    startTurningLeft();
	    break;
	  case TURN_RIGHT:
	    startTurningRight();
	    break;
	  default:
	    stopTurning();
	}
    }

    protected void infoLoggedIntoServer() {
	/* when this method is called, everything is set up for us. */
	setName("R:Aim");
    }

    /* we want to know when the user is turning manually, so we
     * override the callback methods that get called when the user
     * does so. */
    protected void infoManualStartTurningLeft() {
	manualTurnLeft = true;
    }

    protected void infoManualStopTurningLeft() {
	manualTurnLeft = false;
    }

    protected void infoManualStartTurningRight() {
	manualTurnRight = true;
    }

    protected void infoManualStopTurningRight() {
	manualTurnRight = false;
    }

    /* we also want a way to disable/enable the aiming feature. this
     * method gets called whenever the user presses a key that has no
     * special meaning in the game. we pick a key (Z), and check for
     * it. */
    protected void infoUnhandeledKeyPressed(int code) {
	if (code == KeyEvent.VK_Z) {
	    disabled = !disabled;
	    showMessage("Autmatic aiming "
			+ (disabled ? "disabled" : "enabled") + ".");
	}
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    public Aim() {
	/* can't do much in the constructor, as the connection to the
         * client etc. isn't made yet. set up my private variables
         * only. initial queries on the game may be done in the method
         * infoLoggedIntoServer, which is called after a successful
         * connect to the server, and after the server has sent the
         * world info. */
	disabled = false;
    }
}
